# -*- coding: utf-8 -*-
"""Testcase for the magdata module."""

import os
import unittest
import pytest

from numbers import Number

import numpy as np
import numpy.testing

# from pyramid.fielddata import VectorData
# from pyramid import load_vectordata
from empyre.fields import Field


def assert_allclose(actual, desired, rtol=1e-07, atol=1e-08, equal_nan=True, err_msg='', verbose=True):
    return numpy.testing.assert_allclose(
        actual=actual,
        desired=desired,
        rtol=rtol,
        atol=atol,
        equal_nan=equal_nan,
        err_msg=err_msg,
        verbose=verbose,
    )


@pytest.fixture
def fielddata_path():
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test_fielddata')


@pytest.fixture
def vector_data():
    magnitude = np.zeros((4, 4, 4, 3))
    magnitude[1:-1, 1:-1, 1:-1] = 1
    return Field(magnitude, 10.0, vector=True)


@pytest.fixture
def vector_data_asymm():
    shape = (5, 7, 11, 3)
    data = np.linspace(0, 1, np.prod(shape))
    return Field(data.reshape(shape), 10.0, vector=True)


@pytest.fixture
def vector_data_asymmcube():
    shape = (3, 3, 3, 3)
    data = np.linspace(0, 1, np.prod(shape))
    return Field(data.reshape(shape), 10.0, vector=True)


@pytest.fixture
def scalar_data():
    magnitude = np.zeros((4, 4, 4))
    magnitude[1:-1, 1:-1, 1:-1] = 1
    return Field(magnitude, 10.0, vector=False)


def test_copy(vector_data):
    vector_data = vector_data.copy()
    # Make sure it is a new object
    assert vector_data != vector_data, 'Unexpected behaviour in copy()!'
    assert np.allclose(vector_data, vector_data)


def test_bin(vector_data):
    binned_data = vector_data.bin(2)
    reference = 1 / 8. * np.ones((2, 2, 2, 3))
    assert_allclose(binned_data, reference,
                    err_msg='Unexpected behavior in scale_down()!')
    assert_allclose(binned_data.scale, (20, 20, 20),
                    err_msg='Unexpected behavior in scale_down()!')


def test_zoom(vector_data):
    zoomed_test = vector_data.zoom(2, order=0)
    reference = np.zeros((8, 8, 8, 3))
    reference[2:6, 2:6, 2:6] = 1
    assert_allclose(zoomed_test, reference,
                    err_msg='Unexpected behavior in zoom()!')
    assert_allclose(zoomed_test.scale, (5, 5, 5),
                    err_msg='Unexpected behavior in zoom()!')


@pytest.mark.parametrize(
    'mode', [
        'constant',
        'edge',
        'wrap'
    ]
)
@pytest.mark.parametrize(
    'pad_width,np_pad', [
        (1, ((1, 1), (1, 1), (1, 1), (0, 0))),
        ((1, 2, 3), ((1, 1), (2, 2), (3, 3), (0, 0))),
        (((1, 2), (3, 4), (5, 6)), ((1, 2), (3, 4), (5, 6), (0, 0)))
    ]
)
def test_pad(vector_data, mode, pad_width, np_pad):
    magdata_test = vector_data.pad(pad_width, mode=mode)
    reference = np.pad(vector_data, np_pad, mode=mode)
    assert_allclose(magdata_test, reference,
                    err_msg='Unexpected behavior in pad()!')


@pytest.mark.parametrize(
    'axis', [-1, 3]
)
def test_component_reduction(vector_data, axis):
    # axis=-1 is supposed to reduce over the component dimension, if it exists. axis=3 should do the same here!
    res = np.sum(vector_data, axis=axis)
    ref = np.zeros((4, 4, 4))
    ref[1:-1, 1:-1, 1:-1] = 3
    assert res.shape == ref.shape, 'Shape mismatch!'
    assert_allclose(res, ref, err_msg="Unexpected behavior of axis keyword")
    assert isinstance(res, Field), 'Result is not a Field object!'
    assert not res.vector, 'Result is a vector field, but should be reduced to a scalar!'


@pytest.mark.parametrize(
    'axis', [(0, 1, 2), (2, 1, 0), None, (-4, -3, -2)]
)
def test_full_reduction(vector_data, axis):
    res = np.sum(vector_data, axis=axis)
    ref = np.zeros((3,))
    ref[:] = 8
    assert res.shape == ref.shape
    assert_allclose(res, ref, err_msg="Unexpected behavior of full or default reduction")
    assert isinstance(res, np.ndarray)


@pytest.mark.parametrize(
    'axis', [-1, 2]
)
def test_last_reduction_scalar(scalar_data, axis):
    # axis=-1 is supposed to reduce over the component dimension if it exists.
    # In this case it doesn't!
    res = np.sum(scalar_data, axis=axis)
    ref = np.zeros((4, 4))
    ref[1:-1, 1:-1] = 2
    assert res.shape == ref.shape
    assert_allclose(res, ref, err_msg="Unexpected behavior of axis keyword")
    assert isinstance(res, Field)
    assert not res.vector


@pytest.mark.parametrize(
    'axis', [(0, 1, 2), (2, 1, 0), None, (-1, -2, -3)]
)
def test_full_reduction_scalar(scalar_data, axis):
    res = np.sum(scalar_data, axis=axis)
    ref = 8
    assert res.shape == ()
    assert_allclose(res, ref, err_msg="Unexpected behavior of full or default reduction")
    assert isinstance(res, Number)


def test_binary_operator_vector_number(vector_data):
    res = vector_data + 1
    ref = np.ones((4, 4, 4, 3))
    ref[1:-1, 1:-1, 1:-1] = 2
    assert res.shape == ref.shape
    assert_allclose(res, ref, err_msg="Unexpected behavior of addition")
    assert isinstance(res, Field)
    assert res.vector


def test_binary_operator_vector_scalar(vector_data, scalar_data):
    res = vector_data + scalar_data
    ref = np.zeros((4, 4, 4, 3))
    ref[1:-1, 1:-1, 1:-1] = 2
    assert res.shape == ref.shape
    assert_allclose(res, ref, err_msg="Unexpected behavior of addition")
    assert isinstance(res, Field)
    assert res.vector


def test_binary_operator_vector_vector(vector_data):
    res = vector_data + vector_data
    ref = np.zeros((4, 4, 4, 3))
    ref[1:-1, 1:-1, 1:-1] = 2
    assert res.shape == ref.shape
    assert_allclose(res, ref, err_msg="Unexpected behavior of addition")
    assert isinstance(res, Field)
    assert res.vector


@pytest.mark.xfail
def test_binary_operator_vector_broadcast(vector_data):
    # Broadcasting between vector fields is currently not implemented
    second = np.zeros((4, 4, 3))
    second[1:-1, 1:-1] = 1
    second = Field(second, 10.0, vector=True)
    res = vector_data + second
    ref = np.zeros((4, 4, 4, 3))
    ref[1:-1, 1:-1, 1:-1] = 1
    ref[:, 1:-1, 1:-1] += 1
    assert res.shape == ref.shape
    assert_allclose(res, ref, err_msg="Unexpected behavior of addition")
    assert isinstance(res, Field)
    assert res.vector


def test_mask(vector_data):
    mask = vector_data.mask
    reference = np.zeros((4, 4, 4))
    reference[1:-1, 1:-1, 1:-1] = True
    assert_allclose(mask, reference,
                    err_msg='Unexpected behavior in mask attribute!')


def test_get_vector(vector_data):
    mask = vector_data.mask
    vector = vector_data.get_vector(mask)
    reference = np.ones(np.sum(mask) * 3)
    assert_allclose(vector, reference,
                    err_msg='Unexpected behavior in get_vector()!')


def test_set_vector(vector_data):
    mask = vector_data.mask
    vector = 2 * np.ones(np.sum(mask) * 3)
    vector_data.set_vector(vector, mask)
    reference = np.zeros((4, 4, 4, 3))
    reference[1:-1, 1:-1, 1:-1] = 2
    assert_allclose(vector_data, reference,
                    err_msg='Unexpected behavior in set_vector()!')


def test_flip(vector_data_asymm):
    field_flipz = vector_data_asymm.flip(0)
    field_flipy = vector_data_asymm.flip(1)
    field_flipx = vector_data_asymm.flip(2)
    field_flipxy = vector_data_asymm.flip((1, 2))
    field_flipdefault = vector_data_asymm.flip()
    field_flipcomp = vector_data_asymm.flip(-1)
    assert_allclose(np.flip(vector_data_asymm.data, axis=0) * [1, 1, -1], field_flipz.data,
                    err_msg='Unexpected behavior in flip()! (z)')
    assert_allclose(np.flip(vector_data_asymm.data, axis=1) * [1, -1, 1], field_flipy.data,
                    err_msg='Unexpected behavior in flip()! (y)')
    assert_allclose(np.flip(vector_data_asymm.data, axis=2) * [-1, 1, 1], field_flipx.data,
                    err_msg='Unexpected behavior in flip()! (x)')
    assert_allclose(np.flip(vector_data_asymm.data, axis=(1, 2)) * [-1, -1, 1], field_flipxy.data,
                    err_msg='Unexpected behavior in flip()! (xy)')
    assert_allclose(np.flip(vector_data_asymm.data, axis=(0, 1, 2)) * [-1, -1, -1], field_flipdefault.data,
                    err_msg='Unexpected behavior in flip()! (default)')
    assert_allclose(np.flip(vector_data_asymm.data, axis=-1) * [1, 1, 1], field_flipcomp.data,
                    err_msg='Unexpected behavior in flip()! (components)')


@pytest.mark.parametrize(
    'axis', ['x', 'y', 'z']
)
def test_rotate_360(vector_data_asymm, axis):
    assert_allclose(vector_data_asymm.rot90(axis=axis).rot90(axis=axis).rot90(axis=axis).rot90(axis=axis), vector_data_asymm,
                    err_msg=f'Unexpected behavior in rot()! {axis}')


@pytest.mark.parametrize(
    'rot_axis,flip_axes', [
        ('x', (0, 1)),
        ('y', (0, 2)),
        ('z', (1, 2))
    ]
)
def test_rotate_180(vector_data_asymm, rot_axis, flip_axes):
    res = vector_data_asymm.rot90(axis=rot_axis).rot90(axis=rot_axis)
    ref = vector_data_asymm.flip(axis=flip_axes)
    assert_allclose(res, ref, err_msg=f'Unexpected behavior in rot90()! {rot_axis}')


@pytest.mark.parametrize(
    'rot_axis', [
        'x',
        'y',
        'z',
    ]
)
def test_rotate_compare_rot90_1(vector_data_asymmcube, rot_axis):
    
    res = vector_data_asymmcube.rotate(angle=90, axis=rot_axis)
    ref = vector_data_asymmcube.rot90(axis=rot_axis)
    print("input", vector_data_asymmcube.data)
    print("ref", res.data)
    print("res", ref.data)
    assert_allclose(res, ref, err_msg=f'Unexpected behavior in rotate()! {rot_axis}')
